##==============================================================================
##
##  crt0.S
##
##  nds32 startup code
##
##==============================================================================
##
## Copyright (c) 2012-2015 Andes Technology Corporation
## All rights reserved.
##

#include <nds32_init.inc>
#include "../config.h"
#include "interrupt.h"

##------------------------------------------------------------------------------
	.macro vector name
	.align 2
	j OS_Trap_\name
	.endm

	.macro OS_Trap_Interrupt_HW num, nested
OS_Trap_Interrupt_HW\num:
	push $r0
	li  $r0, \num
	.if (\nested == 1)
	b   common_nested_ISR_wrapper
	.else
	b   common_ISR_wrapper
	.endif
	.endm

	.macro LED x
// Disable LED display because the simulator doesn't support LED function.
#ifndef	CFG_SIMU
	hal_set_led \x
#endif
	.endm

	.section .nds32_init, "ax"
!========================================================================
! Vector table
!========================================================================
	.align 2
exception_vector:
	j _start                    !  (0) Trap Reset
	vector TLB_Fill             !  (1) Trap TLB fill
	vector PTE_Not_Present      !  (2) Trap PTE not present
	vector TLB_Misc             !  (3) Trap TLB misc
	vector TLB_VLPT_Miss        !  (4) Trap TLB VLPT miss
	vector Machine_Error        !  (5) Trap Machine error
	vector Debug_Related        !  (6) Trap Debug related
	vector General_Exception    !  (7) Trap General exception
	vector Syscall              !  (8) Syscall

	/* HW interrupt vector layout */
	hal_hw_vectors

	.align  2
exception_vector_end:

ISR_TABLE:
	.long HW0_ISR
	.long HW1_ISR
	.long HW2_ISR
	.long HW3_ISR
	.long HW4_ISR
	.long HW5_ISR
	.long HW6_ISR
	.long HW7_ISR
	.long HW8_ISR
	.long HW9_ISR
	.long HW10_ISR
	.long HW11_ISR
	.long HW12_ISR
	.long HW13_ISR
	.long HW14_ISR
	.long HW15_ISR
	.long HW16_ISR
	.long HW17_ISR
	.long HW18_ISR
	.long HW19_ISR
	.long HW20_ISR
	.long HW21_ISR
	.long HW22_ISR
	.long HW23_ISR
	.long HW24_ISR
	.long HW25_ISR
	.long HW26_ISR
	.long HW27_ISR
	.long HW28_ISR
	.long HW29_ISR
	.long HW30_ISR
	.long HW31_ISR
/*
 * exception handlers
 */

	.section .text

OS_Trap_TLB_Fill:
OS_Trap_PTE_Not_Present:
OS_Trap_TLB_Misc:
OS_Trap_TLB_VLPT_Miss:
OS_Trap_Machine_Error:
OS_Trap_Debug_Related:
OS_Trap_General_Exception:
1:  b   1b

OS_Trap_Syscall:
	SYSCALL_SAVE_ALL
	bal syscall_handler
	SYSCALL_RESTORE_ALL
	iret

common_nested_ISR_wrapper:
	IRQ_NESTED_SAVE_ALL
	la  $r1, ISR_TABLE
	lw  $r1, [$r1+$r0<<2]
	jral $r1
	IRQ_NESTED_RESTORE_ALL
	iret

common_ISR_wrapper:
	IRQ_SAVE_ALL
	la  $r1, ISR_TABLE
	lw  $r1, [$r1+$r0<<2]
	jral $r1
	IRQ_RESTORE_ALL
	iret

OS_Trap_Interrupt_UNDEF:
	b   OS_Trap_Interrupt_UNDEF

	/* Other HW interrupt handlers */
	hal_hw_ISR

##------------------------------------------------------------------------------
## Startup code

#ifdef CFG_LLINIT
#ifdef CFG_REMAP
#define LED_VALUE (0x66) //Burn
#else
#define LED_VALUE (0x11) //xIp
#endif
#else
#define LED_VALUE (0x00) //lOad
#endif

	.global _start
	.func _start
	.type _start, @function

	.align 2
_start:
	!************************** Begin of do-not-modify **************************
	! Please don't modify this code
	! Initialize the registers used by the compiler
	nds32_init              ! NDS32 startup initial macro in <nds32_init.inc>

#ifndef __NDS32_ISA_V3M__
_i:
	mfsr  $r0, $ICM_CFG     ! $r0 = ICM_CFG
	andi  $r0, $r0, 0x1c0   ! $r0 = $r0 & 0x1c0 , take ISZ
	beqz  $r0, _d           ! if no I-Cache, jump to _d

	mfsr  $r0, $CACHE_CTL
	bclr  $r0, $r0, 0
	mtsr  $r0, $CACHE_CTL	! disable I cache

_d:
	mfsr  $r1, $DCM_CFG     ! $r1 = DCM_CFG
	andi  $r1, $r1, 0x1c0   ! $r1 = $r1 & 0x1c0 , take DSZ
	beqz  $r1, _finish      ! if no D-Cache, jump to _finish

	mfsr  $r1, $CACHE_CTL
	bclr  $r1, $r1, 1
	mtsr  $r1, $CACHE_CTL	! disable D cache

_finish:
#endif

!*************************** End of do-not-modify ***************************

	LED LED_VALUE           ! light LED
#ifdef CFG_LLINIT
	bal _nds32_init_mem
#endif
	bal __init
	bal main
1:  b   1b

	.size _start, .-_start

	.end
